/* -*- mode: c++; c-file-style: "engine"; c-basic-offset: 4; indent-tabs-mode: nil -*- */
/****************************************************************************
 * Copyright 2012
 *

 *
 *
 *
 * This software is a computer program whose purpose is to [describe
 * functionalities and technical features of your software].
 * 
 * This software is governed by the CeCILL  license under French law and
 * abiding by the rules of distribution of free software.  You can  use, 
 * modify and/ or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info". 
 * 
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability. 
 * 
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or 
 * data to be ensured and,  more generally, to use and operate it in the 
 * same conditions as regards security. 
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 ****************************************************************************/
/**
 * @file
 * @brief 
 * This file contains the supporting library of the component generator for the C++ language. 
 *
 * This file contains the supporting library of the component generator for 
 * the C++ language. Each component generated using CXXGenerator rests on this 
 * header only library to build components. This file allows for a better 
 * separation between the generic code generated by the component and the 
 * common api to develop components. 
 *  
 * 2007-01-30
 * Olivier Delannoy
 *
 *
 */
#ifndef YML_XMP_COMPONENT_BASE_HH
#define YML_XMP_COMPONENT_BASE_HH 1


#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <mpi.h>


typedef struct{
	int size[5];
	int dataDim;
	int nodes[5];
	int nodesDim;
	int distridubtion[5];
	char aligning[5];
}Map_Info_t;


void parse_align(char * alignDescription, char * align)
{
	char *last=NULL;
	char *tok;
	int i;

	tok=strtok(alignDescription, ":");

	for(tok=strtok_r(tok, "[]", &last),i=0; tok; tok=strtok_r(NULL, "[]", &last),i++)
	{
		align[i]=tok[0];
	}

}

void parse_mapping(Map_Info_t* mapping, char * mapping_description)
{
	int i=0,j=0;
	int len;
	char *work, *tok, *lastWork, *lastTok;
	int * curData, *curSize;

	len=strlen(mapping_description)+1;
	work=(char*)malloc(len*sizeof(char));
	strcpy(work,mapping_description);

	for(i=0; i < 5; i++)
	{
		mapping->size[i]=1;
		mapping->nodes[i]=1;
	}


	for(tok=strtok_r(work,";",&lastWork),j=0; tok; tok=strtok_r(NULL,";",&lastWork),j++)
	{
		switch(j)
		{
		case 0:
			curData=mapping->size;
			curSize=&mapping->dataDim;
			break;
		case 1:
			curData=mapping->nodes;
			curSize=&mapping->nodesDim;
			break;
		case 2:
			curData=mapping->distridubtion;
			break;
		case 3:
			parse_align(tok,mapping->aligning);
			break;
		}

		if( j < 3)
		{
			for(tok=strtok_r(tok,",",&lastTok),i=0; tok; tok=strtok_r(NULL,",",&lastTok),i++)
			{
				*curData=strtol(tok,NULL,0);
				curData++;

			}

			if( j < 2) *curSize=i;
		}

	}

	/* Temporary Solution for changing the * dim of nodes */
	for(i=0;i < 4; i++)
	{
		if(mapping->aligning[i]=='*'&& i==0)
		{
			j=mapping->nodes[i];
			mapping->nodes[i+1] = j;
			mapping->nodes[i] = 1;

		}
	}


	free(work);

}


bool type_import(void * param,char* filename,MPI_Datatype (*ptFuncType)(void), bool (*ptFuncImport)(void*,const char*,const MPI_Datatype, const int), char* mapping_description)
{
	int size=1,i;
	Map_Info_t mapping;
	bool status_import;
	int rank;
	int lsize[2];
	int start[2];
	MPI_Datatype motif;

	parse_mapping(&mapping,mapping_description);
	MPI_Comm_rank(MPI_COMM_WORLD,&rank);

	if(mapping.dataDim > mapping.nodesDim)
	{


		for(i=0; i < mapping.dataDim; i++)
		{
			lsize[i] = mapping.size[i] / mapping.nodes[i];
			size  *= lsize[i];

			if(mapping.aligning[i]=='*')
			{
				start[i] = 0;
			}
			else
			{

				start[i] = rank*lsize[i];
			}

		}


	}
	else if(mapping.dataDim == mapping.nodesDim)
	{

		for(i=0; i < mapping.dataDim; i++)
		{
			lsize[i] = mapping.size[i] / mapping.nodes[i];
			size *= lsize[i];
		}

		if( mapping.dataDim < 2)
		{
			start[0]=(rank/mapping.nodes[0])*lsize[0];
		}
		else
		{
			start[0]=(rank/mapping.nodes[0])*lsize[0];
			start[1]=(rank%mapping.nodes[1])*lsize[1];
		}
	}



	MPI_Type_create_subarray(mapping.dataDim,mapping.size,lsize,start,MPI_ORDER_C,ptFuncType(),&motif);
	MPI_Type_commit(&motif);

	status_import= (*ptFuncImport) (param,filename,motif,size);


	MPI_Type_free(&motif);

	return status_import;
}

/* Note:
 * Muliple Dimension are not handled
 * 2d template align on a single array is not supported
 */

bool type_export(void * param, char* filename,MPI_Datatype (*ptFuncType)(void), bool (*ptFuncExport)(const void*, char*,const MPI_Datatype,const int), char* mapping_description)
{
	int i;
	int size=1;
	bool status_export;
	FILE *fp;
	Map_Info_t mapping;
	int rank;
	int lsize[2];
	int start[2];
	MPI_Datatype motif;

	parse_mapping(&mapping,mapping_description);
	MPI_Comm_rank(MPI_COMM_WORLD,&rank);

	if(mapping.dataDim > mapping.nodesDim)
	{


		for(i=0; i < mapping.dataDim; i++)
		{
			lsize[i] = mapping.size[i] / mapping.nodes[i];
			size  *= lsize[i];

			if( mapping.aligning[i]=='*')
			{
				start[i] = 0;

			}
			else
			{

				start[i] = rank*lsize[i];
			}

//			printf("Parsing Nodes=%i  Dim=%i  Size=%i  Dim=%i Align=%s\n", mapping.nodes[i],mapping.nodesDim, mapping.size[i], mapping.dataDim, mapping.aligning);
		}


	}
	else if(mapping.dataDim == mapping.nodesDim)
	{

		for(i=0; i < mapping.dataDim; i++)
		{
			lsize[i] = mapping.size[i] / mapping.nodes[i];
			size *= lsize[i];
		}

		if( mapping.dataDim < 2)
		{
			start[0]=(rank/mapping.nodes[0])*lsize[0];
		}
		else
		{
			start[0]=(rank/mapping.nodes[0])*lsize[0];
			start[1]=(rank%mapping.nodes[1])*lsize[1];
		}
	}



//	printf("Rank %i Start %i %i\n",rank, start[0],start[1] );


	MPI_Type_create_subarray(mapping.dataDim,mapping.size,lsize,start,MPI_ORDER_C,ptFuncType(),&motif);
	MPI_Type_commit(&motif);

	status_export=(*ptFuncExport) (param,filename,motif,size);

	MPI_Type_free(&motif);

	if(rank==0)
	{
		fp=fopen("output.txt","a");
		fprintf(fp,"%s\n",filename);
		fclose(fp);
	}

	return status_export;

}


#endif
